"""Permission storage model."""
import logging
from enum import IntEnum
from typing import List, Optional, Union

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, User
from django.core.exceptions import ValidationError
from django.db import models

logger = logging.getLogger(__name__)


class Permission(IntEnum):
    """Permission values.

    There is no need for permission to have its own bit since permissions with
    higher numbers always include all lower permissions.
    """

    VIEW = 1
    EDIT = 2
    SHARE = 4
    OWNER = 8

    def has_permission(self, permission: "Permission") -> bool:
        """Return True if the given permission in included in current one."""
        return self >= permission

    @staticmethod
    def from_name(permission_name: str) -> "Permission":
        """Get the permission from name."""
        mapping = {
            "view": Permission.VIEW,
            "edit": Permission.EDIT,
            "share": Permission.SHARE,
            "owner": Permission.OWNER,
        }
        return mapping[permission_name]

    @staticmethod
    def highest_permission() -> "Permission":
        """Return the highest permission."""
        return Permission.OWNER

    def all_permissions(self) -> List[str]:
        """Get a list of all permissions (as strings) included in this one."""
        return [
            str(permission)
            for permission in Permission
            if permission.value <= self.value
        ]

    def all_permission_objects(self) -> List["Permission"]:
        """Get a list of all permissions (as objects) included in this one.

        The permissions are sorted from low to high.
        """
        return sorted(
            [permission for permission in Permission if permission.value <= self.value]
        )

    def __str__(self) -> str:
        """Get the string representation."""
        mapping = {
            Permission.VIEW: "view",
            Permission.EDIT: "edit",
            Permission.SHARE: "share",
            Permission.OWNER: "owner",
        }
        return mapping[self]


class PermissionModel(models.Model):
    """Permission on the permission group."""

    permission = models.PositiveSmallIntegerField()
    users = models.ManyToManyField(get_user_model(), related_name="permission_model")
    groups = models.ManyToManyField(Group(), related_name="permission_model")
    permission_group = models.ForeignKey(
        "PermissionGroup", on_delete=models.CASCADE, related_name="permissions"
    )

    def __str__(self) -> str:
        """Get the string representation."""
        return f"PermissionModel({self.id}, {Permission(self.permission)}, {self.users.all()}, {self.groups.all()})"


class PermissionGroup(models.Model):
    """Group of objecs sharing the same permission."""

    def __str__(self) -> str:
        """Get the string representation."""
        return f"PermissionGroup({self.id}, {self.permissions.all()})"


class PermissionObject(models.Model):
    """Base permission object.

    Every object that has permissions must inherit from this one.
    """

    class Meta:
        """PermissionObject meta options."""

        abstract = True

    #: permission group for the object
    permission_group = models.ForeignKey(
        PermissionGroup, on_delete=models.CASCADE, related_name="%(class)s", null=True
    )

    def __init__(self, *args, **kwargs):
        """Initialize."""
        # The properties used to determine when object is in the container. The
        # order of properties is important: properties listed sooner have
        # precedence over properties listed later when inheriting permission
        # group.
        self._container_properties = ("collection", "entity")
        self._container_attributes = dict()
        super().__init__(*args, **kwargs)
        self._refresh_container_attributes()

    def _refresh_container_attributes(self):
        """Refresh values of container properties.

        Store them into private attributes to avoid hitting the database each
        time we need them.
        """
        self._container_attributes = {
            container_property: getattr(self, f"{container_property}_id", None)
            for container_property in self._container_properties
        }

    def set_permission(
        self, permission: Optional[str], user_or_group: Union[User, Group]
    ):
        """Enable use of syntax obj.set_permission instead."""
        # Resolve circular import.
        from resolwe.permissions.utils import set_permission

        set_permission(permission, user_or_group, self)

    @property
    def topmost_container(self) -> Optional[models.Model]:
        """Get the top-most container.

        :returns: the top-most container or None if it does not exist.
        """
        for property in self._container_properties:
            value = getattr(self, property)
            if value is not None:
                return value
        return None

    def in_container(self, cached: bool = True) -> bool:
        """Return if object is contained in a container."""
        if cached:
            return any(self._container_attributes.values())
        else:
            return any(
                getattr(self, f"{container_property}_id", False)
                for container_property in self._container_properties
            )

    def save(self, *args, **kwargs):
        """Override the save method.

        When 'permission_group' attribute is not set create one. This could
        lead to a lot unassigned permission groups lying around.
        """
        # When object is contaied in the container it inherits the permission
        # group from the container. The outermost container has precedence.
        if self.in_container(cached=False):
            for container_property in self._container_properties:
                value = getattr(self, f"{container_property}_id", None)
                # Object was moved to a different container.
                if value != self._container_attributes[container_property]:
                    container = getattr(self, container_property)
                    self.permission_group_id = container.permission_group_id
                    # Add collection tags to data objects if they are moved to
                    # a new collection.
                    if (
                        self._meta.label in ["flow.Data", "flow.Entity"]
                        and container_property == "collection"
                    ):
                        self.tags = container.tags

                    # When moving entity all data objects in the entity must
                    # have the permission group set to the new one and tags
                    # updated.
                    if self._meta.label == "flow.Entity":
                        # Update permission group and tags.
                        self.data.update(
                            permission_group_id=container.permission_group_id
                        )
                        self.data.update(tags=container.tags)
                        self.data.update(collection_id=value)
                    break
                # Is this if necessary?
                # Object was moved to a different container.
                elif value is not None:
                    container = getattr(self, container_property)
                    self.permission_group_id = container.permission_group_id
                    break

            assert (
                self.permission_group_id is not None
            ), f"Container object of {self}({self._meta.label}) must have permission group set."
        elif self.in_container(cached=True):
            raise ValidationError("Object can only be moved to another container.")
        elif self.permission_group is None:
            self.permission_group = PermissionGroup.objects.create()

        super().save(*args, **kwargs)
        self._refresh_container_attributes()



# Generated by Django 3.1.7 on 2021-10-12 11:34

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0002_squashed_0049_json_field"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="collection",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view collection"),
                    ("edit", "Can edit collection"),
                    ("share", "Can share collection"),
                    ("owner", "Is owner of the collection"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="data",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view data"),
                    ("edit", "Can edit data"),
                    ("share", "Can share data"),
                    ("owner", "Is owner of the data"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="descriptorschema",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view descriptor schema"),
                    ("edit", "Can edit descriptor schema"),
                    ("share", "Can share descriptor schema"),
                    ("owner", "Is owner of the description schema"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="entity",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view entity"),
                    ("edit", "Can edit entity"),
                    ("share", "Can share entity"),
                    ("owner", "Is owner of the entity"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="process",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view process"),
                    ("share", "Can share process"),
                    ("owner", "Is owner of the process"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="relation",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view relation"),
                    ("edit", "Can edit relation"),
                    ("share", "Can share relation"),
                    ("owner", "Is owner of the relation"),
                ),
            },
        ),
    ]



""".. Ignore pydocstyle D400.

==================
Permissions Filter
==================

"""
from rest_framework.filters import BaseFilterBackend

from resolwe.permissions.shortcuts import get_objects_for_user


class ResolwePermissionsFilter(BaseFilterBackend):
    """Permissions filter."""

    def filter_queryset(self, request, queryset, view):
        """Filter permissions queryset."""
        user = request.user

        app_label = queryset.model._meta.app_label
        permission = f"{app_label}.view"
        return get_objects_for_user(user, permission, queryset)



# Generated by Django 3.1.7 on 2021-10-12 11:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("permissions", "0001_initial"),
        ("flow", "0003_auto_20211012_1134"),
    ]

    operations = [
        migrations.AddField(
            model_name="collection",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="collection",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="data",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="data",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="descriptorschema",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="descriptorschema",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="entity",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="entity",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="process",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="process",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="relation",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="relation",
                to="permissions.permissiongroup",
            ),
        ),
    ]



"""Permissions functions used in Resolwe Viewsets."""
from collections import defaultdict
from typing import Optional

from django.conf import settings
from django.db import models, transaction

from rest_framework import exceptions, serializers, status
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.permissions.models import Permission
from resolwe.permissions.shortcuts import get_object_perms, get_user_group_perms_qset

from .utils import (
    check_owner_permission,
    check_public_permissions,
    check_user_permissions,
    update_permission,
)


class CurrentUserPermissionsSerializer(serializers.Serializer):
    """Current user permissions serializer."""

    id = serializers.IntegerField()
    type = serializers.CharField(max_length=50)
    name = serializers.CharField(max_length=100)
    permissions = serializers.ListField(child=serializers.CharField(max_length=30))


class ResolwePermissionsMixin:
    """Mixin to support managing `Resolwe` objects' permissions."""

    def get_serializer_class(self):
        """Augment base serializer class.

        Include permissions information with objects.

        """
        base_class = super().get_serializer_class()

        class SerializerWithPermissions(base_class):
            """Augment serializer class."""

            def __init__(
                serializer_self,
                instance: Optional[models.QuerySet] = None,
                *args,
                **kwargs,
            ):
                """Override init."""
                # print("Instance")
                # print(instance, type(instance), instance.all())
                if isinstance(instance, models.Model):
                    query_set = instance._meta.model.objects.filter(pk=instance.pk)
                # elif isinstance(
                #     instance, (models.QuerySet, models.query.EmptyQuerySet)
                # ):
                else:
                    query_set = instance
                # elif instance is not None:
                #     raise RuntimeError(
                #         f"Instance {instance} must be a QuerySet or Model."
                #     )
                if instance is not None:
                    serializer_self._permissions_qset = get_user_group_perms_qset(
                        self.request.user, query_set
                    )
                serializer_self._permission_map = None
                super().__init__(instance, *args, **kwargs)

            def get_fields(serializer_self):
                """Return serializer's fields."""
                fields = super().get_fields()
                fields["current_user_permissions"] = CurrentUserPermissionsSerializer(
                    read_only=True
                )
                return fields

            def _set_permission_map(serializer_self):
                """Set the permission_map property."""
                if serializer_self._permission_map is None:
                    serializer_self._permission_map = defaultdict(list)
                    for value in serializer_self._permissions_qset:
                        entity_type = "user" if value["is_user"] else "group"
                        if (
                            entity_type == "user"
                            and value["entityname"] == settings.ANONYMOUS_USER_NAME
                        ):
                            entity_type = "public"
                        serializer_self._permission_map[value["pk"]].append(
                            {
                                "type": entity_type,
                                "id": value["entityid"],
                                "name": value["entityname"],
                                "permissions": Permission(
                                    value["permission"]
                                ).all_permissions(),
                            }
                        )

            def create(serializer_self, validated_data):
                """Override create.

                After instance is created from validated data construct
                permissions queryset and store it.
                """
                instance = super().create(validated_data)
                if isinstance(instance, models.Model):
                    query_set = instance._meta.model.objects.filter(pk=instance.pk)
                elif isinstance(instance, models.QuerySet):
                    query_set = instance
                if instance is not None:
                    serializer_self._permissions_qset = get_user_group_perms_qset(
                        self.request.user, query_set
                    )
                return instance

            def to_representation(serializer_self, instance: models.Model):
                """Object serializer."""
                data = super().to_representation(instance)
                if (
                    "fields" not in self.request.query_params
                    or "current_user_permissions" in self.request.query_params["fields"]
                ):
                    serializer_self._set_permission_map()
                    # data["current_user_permissions"] = get_object_perms(
                    #    instance, self.request.user
                    # )
                    data["current_user_permissions"] = serializer_self._permission_map[
                        instance.pk
                    ]

                return data

        return SerializerWithPermissions

    @action(
        detail=True,
        methods=["get", "post"],
        url_path="permissions",
        url_name="permissions",
    )
    def detail_permissions(self, request, pk=None):
        """Get or set permissions API endpoint."""
        # print("Detail permission"*1000)
        obj = self.get_object()

        if request.method == "POST":
            payload = request.data
            user = request.user
            is_owner = user.has_perm("owner", obj=obj)
            allow_owner = is_owner or user.is_superuser
            check_owner_permission(payload, allow_owner)
            check_public_permissions(payload)
            check_user_permissions(payload, request.user.pk)

            with transaction.atomic():
                update_permission(obj, payload)
                owner_permission = obj.permission_group.permissions.filter(
                    permission=Permission.from_name("owner").value
                ).first()
                owner_count = 0
                if owner_permission is not None:
                    owner_count = owner_permission.users.count()

                if not owner_count:
                    raise exceptions.ParseError("Object must have at least one owner.")

        return Response(get_object_perms(obj))

    @action(
        detail=False,
        methods=["get", "post"],
        url_path="permissions",
        url_name="permissions",
    )
    def list_permissions(self, request):
        """Batch get or set permissions API endpoint."""
        # TODO: Implement batch get/set permissions
        return Response(status=status.HTTP_501_NOT_IMPLEMENTED)



# pylint: disable=missing-docstring
from copy import deepcopy

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

from rest_framework import exceptions, status

from resolwe.flow.models import Collection, DescriptorSchema, Process
from resolwe.flow.views import CollectionViewSet, DescriptorSchemaViewSet
from resolwe.permissions.models import Permission, PermissionModel
from resolwe.permissions.utils import (
    assign_contributor_permissions,
    check_owner_permission,
    check_public_permissions,
    check_user_permissions,
    get_anonymous_user,
    get_perms,
    set_permission,
)
from resolwe.test import ResolweAPITestCase, TestCase


class CollectionPermissionsTest(ResolweAPITestCase):
    def setUp(self):
        User = get_user_model()
        self.user1 = User.objects.create(username="test_user1", email="user1@test.com")
        self.user2 = User.objects.create(username="test_user2", email="user2@test.com")
        self.user3 = User.objects.create(username="test_user3", email="user1@test.com")
        self.owner = User.objects.create(username="owner")

        self.public = get_anonymous_user()
        self.group = Group.objects.create(name="Test group")

        self.collection = Collection.objects.create(
            contributor=self.owner, name="Test collection 1"
        )
        self.collection.set_permission("owner", self.owner)

        self.process = Process.objects.create(
            name="Test process",
            contributor=self.owner,
        )

        self.resource_name = "collection"
        self.viewset = CollectionViewSet

        super().setUp()

    def test_public_user(self):
        """Public user cannot create/edit anything"""
        set_permission("share", self.user1, self.collection)

        data = {"public": "view"}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        data = {"public": "edit"}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        data = {"public": "share"}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        data = {"public": "owner"}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_user_without_share(self):
        """User without ``SHARE`` permission cannot do anything"""
        set_permission("edit", self.user1, self.collection)

        # Can not add permissions to users.
        data = {"users": {self.user2.pk: "view"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        # Can not add permissions to groups.
        data = {"users": {self.group.pk: "view"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_user_with_share(self):
        self.collection.set_permission("share", self.user1)

        # Can set permissions to users.
        data = {"users": {self.user2.pk: "view"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(get_perms(self.user2, self.collection), ["view"])

        # Can set permissions to groups.
        data = {"groups": {self.group.pk: "view"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(get_perms(self.group, self.collection), ["view"])

    def test_protect_owner(self):
        """Only owners can modify `owner` permission"""
        self.collection.set_permission("share", self.user1)

        # User with share permission cannot grant ``owner`` permission
        data = {"users": {self.user2.pk: "owner"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertNotIn("owner", get_perms(self.user2, self.collection))
        self.assertFalse(PermissionModel.objects.filter(users=self.user2).exists())

        # User with share permission cannot revoke ``owner`` permission
        data = {"users": {self.user2.pk: "owner"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(PermissionModel.objects.filter(users=self.user2).exists())

        set_permission("owner", self.user1, self.collection)

        # ``owner`` permission cannot be assigned to a group
        data = {"groups": {self.group.pk: "owner"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertFalse(PermissionModel.objects.filter(groups=self.group).exists())

        # User with owner permission can grant ``owner`` permission
        data = {"users": {self.user2.pk: "owner"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(
            get_perms(self.user2, self.collection), ["view", "edit", "share", "owner"]
        )

        # User with owner permission can revoke ``owner`` permission
        data = {"users": {self.user2.pk: "edit"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertFalse(
            PermissionModel.objects.filter(
                users=self.user2, permission=Permission.from_name("owner").value
            ).exists()
        )

        # User with owner permission cannot remove all owners
        data = {"users": {self.user1.pk: "edit", self.owner.pk: "edit"}}

        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data["detail"], "Object must have at least one owner.")

        owner_permission = self.collection.permission_group.permissions.get(
            permission=Permission.from_name("owner").value
        )
        owner_count = owner_permission.users.count()
        self.assertEqual(owner_count, 2)

        # User can delete his owner permission if there is at least one other owner
        self.assertTrue(owner_permission.users.filter(pk=self.user1.pk).exists())
        data = {"users": {self.user1.pk: "view"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertFalse(owner_permission.users.filter(pk=self.user1.pk).exists())

    def test_nonexisting_permission(self):
        self.collection.set_permission("share", self.owner)

        # Add one valid permission to make sure that no permission is applied if any of them is unknown.
        data = {"users": {self.user1.pk: "view", self.user2.pk: "foo"}}
        resp = self._detail_permissions(self.collection.pk, data, self.owner)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data["detail"], "Unknown permission: foo")
        self.assertEqual(get_perms(self.user1, self.collection), [])
        self.assertEqual(get_perms(self.user2, self.collection), [])

    def test_nonexisting_user_group(self):
        self.collection.set_permission("share", self.owner)

        # Whole request should fail, so `user1` shouldn't have any permission assigned.
        data = {"users": {"999": "view", self.user1.pk: "view"}}
        resp = self._detail_permissions(self.collection.pk, data, self.owner)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data["detail"], "Unknown user: 999")
        # Maybe different?
        self.assertEqual(get_perms(self.user1, self.collection), [])

        # Whole request should fail, so `group` shouldn't have any permission assigned.
        data = {"groups": {"999": "view", self.group.pk: "view"}}
        resp = self._detail_permissions(self.collection.pk, data, self.owner)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(resp.data["detail"], "Unknown group: 999")
        self.assertEqual(get_perms(self.group, self.collection), [])

        self.collection.set_permission("view", self.user1)
        self.collection.set_permission("view", self.group)

    def test_share_by_email(self):
        set_permission("share", self.user1, self.collection)

        data = {"users": {self.user2.email: "view"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(get_perms(self.user2, self.collection), ["view"])

        # Check if error is raise when trying to share with duplicated email.
        data = {"users": {self.user3.email: "view"}}
        resp = self._detail_permissions(self.collection.pk, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            resp.data["detail"], "Cannot uniquely determine user: user1@test.com"
        )


class DescriptorSchemaPermissionsTest(ResolweAPITestCase):
    def setUp(self):
        self.group = Group.objects.create(name="Test group")

        self.resource_name = "collection"
        self.viewset = DescriptorSchemaViewSet

        super().setUp()

        self.descriptor_schema = DescriptorSchema.objects.create(
            contributor=self.contributor
        )
        assign_contributor_permissions(self.descriptor_schema)

    def test_set_permissions(self):
        # Can add permissions to users.
        data = {"users": {self.user.pk: "view"}}
        resp = self._detail_permissions(
            self.descriptor_schema.pk, data, self.contributor
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertEqual(get_perms(self.user, self.descriptor_schema), ["view"])

        # Can add permissions to groups.
        data = {"groups": {self.group.pk: "view"}}
        resp = self._detail_permissions(
            self.descriptor_schema.pk, data, self.contributor
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(get_perms(self.group, self.descriptor_schema), ["view"])

        # Can remove permissions from users.
        data = {"users": {self.user.pk: "NONE"}}
        resp = self._detail_permissions(
            self.descriptor_schema.pk, data, self.contributor
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(get_perms(self.user, self.descriptor_schema), [])

        # Can remove permissions from groups.
        data = {"groups": {self.group.pk: "NONE"}}
        resp = self._detail_permissions(
            self.descriptor_schema.pk, data, self.contributor
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(get_perms(self.group, self.descriptor_schema), [])


class PermissionsUtilitiesTest(TestCase):
    def test_filter_owner_permission(self):
        """Check that ``owner`` permission is catched everywhere"""
        data_template = {
            "users": {1: "view", 2: "edit", 3: "NONE"},
            "groups": {1: "edit", 2: "NONE"},
        }

        check_owner_permission(data_template, False)

        data = deepcopy(data_template)
        data["users"][1] = "owner"
        with self.assertRaises(exceptions.PermissionDenied):
            check_owner_permission(data, False)
        check_owner_permission(data, True)

        data = deepcopy(data_template)
        data["groups"][1] = "owner"
        with self.assertRaises(exceptions.ParseError):
            check_owner_permission(data, False)
        with self.assertRaises(exceptions.ParseError):
            check_owner_permission(data, True)

    def test_filter_user_permissions(self):
        """Check that user cannot change his own permissions"""
        data = {
            "users": {
                1: "view",
                2: "NONE",
            }
        }

        with self.assertRaises(exceptions.PermissionDenied):
            check_user_permissions(data, 1)

        with self.assertRaises(exceptions.PermissionDenied):
            check_user_permissions(data, 2)

        check_user_permissions(data, 3)

    def test_filter_public_permissions(self):
        """Check that public user cannot get to open permissions"""
        data = {"public": "view"}
        check_public_permissions(data)

        data = {"public": "edit"}
        with self.assertRaises(exceptions.PermissionDenied):
            check_public_permissions(data)

        data = {"public": "share"}
        with self.assertRaises(exceptions.PermissionDenied):
            check_public_permissions(data)

        data = {"public": "owner"}
        with self.assertRaises(exceptions.PermissionDenied):
            check_public_permissions(data)



""".. Ignore pydocstyle D400.

=====================
Permissions shortcuts
=====================

.. autofunction:: _group_groups
.. autofunction:: get_object_perms

"""
from typing import Dict, List, Optional, Tuple, Union

from django.conf import settings
from django.contrib.auth.models import Group, User
from django.db import models

from resolwe.permissions.models import Permission, PermissionModel
from resolwe.permissions.utils import get_anonymous_user, get_identity, user_model_check


def get_user_group_perms_qset(
    user_or_group: Union[User, Group], qset: models.QuerySet
) -> Tuple[List[str], List[Tuple[int, str, List[str]]]]:
    """Get permissions for every object in the queryset."""
    user, groups = get_identity(user_or_group)

    assert user is not None, "User must not be none"

    if groups is None and user is not None:
        groups = user.groups.all()

    assert groups is not None

    filters = models.Q(permission_group__permissions__users=user) | models.Q(
        permission_group__permissions__groups__in=groups.all()
    )

    anonymous_user = get_anonymous_user()
    if user != anonymous_user:
        filters = filters | models.Q(
            permission_group__permissions__users__username=anonymous_user.username
        )

    return (
        qset.filter(filters)
        .annotate(
            entityid=models.functions.Coalesce(
                "permission_group__permissions__users",
                "permission_group__permissions__groups",
            )
        )
        .annotate(
            entityname=models.functions.Coalesce(
                "permission_group__permissions__users__username",
                "permission_group__permissions__groups__name",
            )
        )
        .annotate(
            is_user=models.functions.Coalesce(
                models.functions.Cast(
                    "permission_group__permissions__users",
                    output_field=models.BooleanField(),
                ),
                False,
            )
        )
        .annotate(permission=models.F("permission_group__permissions__permission"))
    ).values("pk", "entityid", "entityname", "is_user", "permission")


def get_user_group_perms(
    user_or_group: Union[User, Group], obj: models.Model
) -> Tuple[List[str], List[Tuple[int, str, List[str]]]]:
    """Get permissins for user groups.

    Based on guardian.core.ObjectPermissionChecker.
    """
    user, groups = get_identity(user_or_group)

    # Inactive users shoud have no permissions.
    if user is not None and not user.is_active:
        return [], []

    user_perms = []
    groups_perms = []

    if user is not None:
        if not user.is_active:
            return [], []
        if user.is_superuser:
            user_perms = Permission.highest_permission().all_permissions()
        else:
            # Get user permissions on obj.
            should_continue, user, obj = user_model_check(user, obj)
            if not should_continue:
                return [], []
            permission = PermissionModel.objects.filter(
                users=user, permission_group=obj.permission_group
            ).values_list("permission", flat=True)
            if permission:
                user_perms = Permission(permission[0]).all_permissions()

        groups = user.groups.all()
    else:
        groups = [groups]

    # Get group permissions for every group
    for group in groups:
        group_qset = PermissionModel.objects.filter(
            groups=group, permission_group=obj.permission_group
        ).values_list("permission", flat=True)
        if group_qset:
            group_permissions = Permission(group_qset[0]).all_permissions()
            groups_perms.append((group.pk, group.name, group_permissions))

    return user_perms, groups_perms


def get_users_with_perms(obj: models.Model) -> Dict[User, List[str]]:
    """Get users with permissions on the given object.

    Only permissions attached to user directly are consider, permissions
    attached to the user groups are ignored.

    :returns: a dict where users are keys and list of permissions are values.

    TODO: Sanity check if object has permission group?
    """
    permissions = {}
    for permission_model in PermissionModel.objects.filter(
        permission_group=obj.permission_group
    ):
        all_permissions = Permission(permission_model.permission).all_permissions()
        for user in permission_model.users.all():
            permissions[user] = all_permissions
    return permissions


def get_groups_with_perms(obj: models.Model) -> Dict[Group, List[str]]:
    """Get groups with permissions on the given object.

    :returns: a dict where groups are keys and list of permissions are values.

    TODO: Sanity check if object has permission group?
    """
    permissions = {}
    for permission_model in PermissionModel.objects.filter(
        permission_group=obj.permission_group
    ):
        all_permissions = Permission(permission_model.permission).all_permissions()
        for group in permission_model.groups.all():
            permissions[group] = all_permissions
    return permissions


def get_object_perms(obj: models.Model, user: Optional[User] = None):
    """Return permissions for given object in Resolwe specific format.

    Function returns permissions for given object ``obj`` in following
    format::

       {
           "type": "group"/"user"/"public",
           "id": <group_or_user_id>,
           "name": <group_or_user_name>,
           "permissions": [<first_permission>, <second_permission>,...]
       }

    For ``public`` type ``id`` and ``name`` keys are omitted.

    If ``user`` parameter is given, permissions are limited only to
    given user, groups he belongs to and public permissions.

    :param obj: Resolwe's DB model's instance
    :type obj: a subclass of :class:`~resolwe.flow.models.base.BaseModel`
    :param user: Django user
    :type user: :class:`~django.contrib.auth.models.User` or :data:`None`
    :return: list of permissions object in described format
    :rtype: list

    """
    perms_list = []
    public_permissions = []

    if user is not None:
        if user.is_authenticated:
            user_perms, groups_perms = get_user_group_perms(user, obj)
        else:
            user_perms, groups_perms = [], []

        if user_perms:
            perms_list.append(
                {
                    "type": "user",
                    "id": user.pk,
                    "name": user.get_full_name() or user.username,
                    "username": user.username,
                    "permissions": user_perms,
                }
            )

        if groups_perms:
            for group_id, group_name, permissions in groups_perms:
                perms_list.append(
                    {
                        "type": "group",
                        "id": group_id,
                        "name": group_name,
                        "permissions": permissions,
                    }
                )
    else:
        for user, permissions in get_users_with_perms(obj).items():
            if user.username == settings.ANONYMOUS_USER_NAME:
                # public user permissions are considered bellow
                public_permissions = permissions
                continue
            perms_list.append(
                {
                    "type": "user",
                    "id": user.pk,
                    "name": user.get_full_name() or user.username,
                    "username": user.username,
                    "permissions": permissions,
                }
            )

        for group, permissions in get_groups_with_perms(obj).items():
            perms_list.append(
                {
                    "type": "group",
                    "id": group.pk,
                    "name": group.name,
                    "permissions": permissions,
                }
            )
    # Retrieve public permissions only if they are not set.
    if not public_permissions:
        public_permissions = get_user_group_perms(get_anonymous_user(), obj)[0]
    if public_permissions:
        perms_list.append({"type": "public", "permissions": public_permissions})

    return perms_list


# based on guardian.shortcuts.get_objects_for_user
def get_objects_for_user(
    user: User,
    perms: Union[List[str], str],
    queryset: models.QuerySet,
    use_groups=True,
    any_perm=False,  # <-- in testing
    with_superuser=True,  # <--
    accept_global_perms=True,
    perms_filter="pk__in",  # <-- REMOVE, no longer needed
) -> models.QuerySet:
    """Return queryset with required permissions."""
    if isinstance(perms, str):
        perms = [perms]

    if not perms:
        return queryset.none()

    app_label = queryset.model._meta.app_label
    codenames = set()

    # Compute codenames set and ctype if possible
    for perm in perms:
        if "." in perm:
            new_app_label, codename = perm.split(".", 1)
            if app_label is not None and app_label != new_app_label:
                raise RuntimeError(
                    "Given perms must have same app label "
                    "({} != {})".format(app_label, new_app_label)
                )
            else:
                app_label = new_app_label
        else:
            codename = perm
        codenames.add(codename)

    # First check if user is superuser and if so, return queryset immediately
    if with_superuser and user.is_superuser:
        return queryset

    # Get anonymous user instance when user in not logged in.
    if user.is_anonymous:
        user = get_anonymous_user()

    # When any permission is sufficient take the bottom, else top one.
    if not any_perm:
        needed_permission = max(Permission.from_name(perm) for perm in codenames).value
    else:
        needed_permission = min(Permission.from_name(perm) for perm in codenames).value

    # Handle special case of Storage and Relation.
    filters_prefix = ""
    if queryset.model._meta.label == "flow.Storage":
        filters_prefix = "data__"
    if queryset.model._meta.label == "flow.Relation":
        filters_prefix = "collection__"

    user_filters = {
        f"{filters_prefix}permission_group__permissions__users": user,
        f"{filters_prefix}permission_group__permissions__permission__gte": needed_permission,
    }
    anonymous_filters = {
        f"{filters_prefix}permission_group__permissions__users": get_anonymous_user(),  # TODO: can this be more efficient?
        f"{filters_prefix}permission_group__permissions__permission__gte": needed_permission,
    }

    if not use_groups:
        ids = (
            queryset.filter(models.Q(**user_filters) | models.Q(**anonymous_filters))
            .distinct()
            .values_list("id", flat=True)
        )
    else:
        group_filters = {
            f"{filters_prefix}permission_group__permissions__groups__in": user.groups.all().values_list(
                "pk", flat=True
            ),
            f"{filters_prefix}permission_group__permissions__permission__gte": needed_permission,
        }

        # Return the original queryset without additional joins.
        # So first filter out all the necessary ids and then firter the
        # original queryset based on the set of ids.

        # Warning: do not remove list here otherwise the query bellow may fail
        # when using ExpressionLateralJoin.
        ids = list(
            queryset.filter(
                models.Q(**user_filters)
                | models.Q(**anonymous_filters)
                | models.Q(**group_filters)
            )
            .distinct()
            .values_list("id", flat=True)
        )

    return queryset.filter(id__in=ids)



""".. Ignore pydocstyle D400.

=================
Permissions utils
=================

.. autofunction:: copy_permissions

"""
from contextlib import suppress
from itertools import chain
from typing import List, Optional, Tuple, Union

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group, User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction

from rest_framework import exceptions

from resolwe.permissions.models import Permission, PermissionModel, PermissionObject


def get_perms(user_or_group: Union[User, Group], obj: models.Model) -> List[str]:
    """Get permissions for user/group on the given object."""
    # TODO: storage, others
    if not model_has_permissions(obj):
        raise RuntimeError(
            f"There is no support for permissions on object of type {obj._meta.label}."
        )
    if not isinstance(obj, PermissionObject):
        raise RuntimeError(f"The given model {obj} has no permission group.")

    # TODO: for storage objects this must be a Data permission group, so a
    # database hit is inevitable.
    permission_group = obj.permission_group
    entity = next(
        entity for entity in get_identity(user_or_group) if entity is not None
    )
    entity_name = "groups" if isinstance(entity, Group) else "users"

    permission_models = permission_group.permissions.filter(**{entity_name: entity})
    assert permission_models.count() <= 1, (
        f"Too many permissions on object of type {obj._meta.label} with "
        f"key {obj.pk} for {entity_name} with key {entity.pk}."
    )

    permission_model = permission_models.first()
    if permission_model is None:
        return []
    calculated_permissions = Permission(permission_model.permission).all_permissions()
    all_object_permissions = get_all_perms(obj)
    return [
        permission
        for permission in calculated_permissions
        if permission in all_object_permissions
    ]


def set_permission(
    perm: Optional[str], user_or_group: Union[User, Group], obj: models.Model
):
    """Set the given permission.

    Since permissions are totally ordered it only makes sense to set the
    permissions. The previous permissions user/group might have on the object
    are discarded.

    When None is given for perm all user permissions are removed.

    :raises AssertionError: when no user or group is given.
    :raises RuntimeError: when given object has no permission_group or is
        contained in a container.
    :raises ValueError: when the given permission can not be found.
    """

    # First perform basic checks on the object itself.
    if not model_has_permissions(obj):
        raise RuntimeError(
            f"There is no support for permissions on object of type {obj._meta.label}."
        )
    if obj.in_container():
        raise RuntimeError(
            f"The permissions can not be set on object {obj} ({obj._meta.label}) in container."
        )
    permission_group = obj.permission_group

    # Now check user/group.
    entity = next(
        entity for entity in get_identity(user_or_group) if entity is not None
    )
    assert entity is not None, "User or group must be given."
    entity_name = "groups" if isinstance(entity, Group) else "users"

    # Remove old permission (if they exist).
    with suppress(PermissionModel.DoesNotExist):
        permission_model = permission_group.permissions.get(**{entity_name: entity})
        getattr(permission_model, entity_name).remove(entity)

    # Assign new permission.
    if perm is not None:
        permission = Permission.from_name(perm)
        new_permission_model = PermissionModel.objects.get_or_create(
            permission=permission.value, permission_group=permission_group
        )[0]
        getattr(new_permission_model, entity_name).add(entity)


def get_user(user: User) -> User:
    """Get the same user or anonymous one when not authenticated.

    :raises django.core.exceptions.ObjectDoesNotExist: when user is not
        authenticated and anonymous user could not be found.
    """
    if user.is_authenticated:
        return user
    else:
        return get_anonymous_user()


def model_has_permissions(obj: models.Model) -> bool:
    """Check whether model has object level permissions."""
    additional_labels = ("flow.Relation", "flow.Storage")
    return hasattr(obj, "permission_group") or obj._meta.label in additional_labels


def user_model_check(user: User, obj: models.Model) -> Tuple[bool, User, models.Model]:
    """Check the user and model and decice how to perform the permission checks.

    :returns: the tuple (should_check, user, object) indicating whether permissions
        should be checked for the returned user.
    """
    # Can only check permissions on specific models.
    if not isinstance(obj, models.Model):
        return False, user, obj

    if not model_has_permissions(obj):
        return False, user, obj

    # Use the anonymous user object if user is not logged in.
    try:
        user = get_user(user)
    except ObjectDoesNotExist:
        return False, user, obj

    return True, user, obj


def get_anonymous_user() -> User:
    """Get the anonymous user.

    Note that is the actual user object with username specified in setting
    ANONYMOUS_USER_NAME or id specified in setting ANONYMOUS_USER_ID. The later
    setting has precedence.

    :raises django.core.exceptions.ObjectDoesNotExist: when anonymous
        user could not be found.
    :raises RuntimeError: when setting ANONYMOUS_USER_NAME could not be found.
    """

    anonymous_user_id = getattr(settings, "ANONYMOUS_USER_ID", None)
    if anonymous_user_id:
        try:
            return get_user_model().objects.get(id=anonymous_user_id)
        except User.DoesNotExist:
            from resolwe.test.utils import is_testing

            # Resolve circular import.
            if is_testing():
                return get_user_model().objects.create(
                    id=anonymous_user_id, username="public", is_active=True, email=""
                )
            else:
                raise

    anonymous_username = getattr(settings, "ANONYMOUS_USER_NAME", None)
    if anonymous_username:
        try:
            return get_user_model().objects.get(
                **{User.USERNAME_FIELD: anonymous_username}
            )
        except User.DoesNotExist:
            # Resolve circular import.
            from resolwe.test.utils import is_testing

            if is_testing():
                return get_user_model().objects.create(
                    username=anonymous_username, is_active=True, email=""
                )
            else:
                raise

    raise RuntimeError("No ANONYMOUS_USER_ID/ANONYMOUS_USER_NAME setting found.")


def get_identity(
    user_group: Union[Group, User]
) -> Tuple[Optional[User], Optional[Group]]:
    """Return the tuple (user, group) where exactly one of them is None.

    :raises RuntimeError: when parameter is neither user nor group instance.
    """
    if isinstance(user_group, AnonymousUser):
        return get_anonymous_user(), None

    UserModel = get_user_model()
    if isinstance(user_group, UserModel):
        return user_group, None

    if isinstance(user_group, Group):
        return None, user_group

    raise RuntimeError("Parameter is not user or group instance.")


def get_all_perms(obj):
    """Return a list of all permissions on ``obj``."""
    return list(zip(*obj._meta.permissions))[0]


def copy_permissions(src_obj: models.Model, dest_obj: models.Model):
    """Copy permissions form ``src_obj`` to ``dest_obj``."""
    # TODO: why is this here? Can this condition ever be fullfilled?
    if src_obj is None or dest_obj is None:
        return

    if not model_has_permissions(src_obj) or not model_has_permissions(dest_obj):
        return

    source_permission_group = src_obj.permission_group
    destination_permission_group = dest_obj.permission_group

    for permission_model in source_permission_group.permissions.all():
        try:
            new_permission_model = PermissionModel.objects.get(
                permission=permission_model.permission,
                permission_group=destination_permission_group,
            )
        except PermissionModel.DoesNotExist:
            new_permission_model = PermissionModel.objects.create(
                permission=permission_model.permission,
                permission_group=destination_permission_group,
            )
        finally:
            # Check for existing permission and assing one only when needed.
            # This could be slow but it is necessary in order to have only one
            # permission perm user/group per permission group, which makes
            # sense if permissions are ordered linearly.
            for entity in chain(
                permission_model.users.all(), permission_model.groups.all()
            ):
                is_user = isinstance(entity, User)
                attribute_name = "users" if is_user else "groups"
                filter = {f"{attribute_name}": entity}
                existing_permissions = destination_permission_group.permissions.filter(
                    **filter
                )
                if existing_permissions:
                    existing_permission = existing_permissions.get()
                    # Higher (or equal) permission is OK.
                    if existing_permission.permission >= permission_model.permission:
                        continue
                    # Lower has to be deleted, then added with the correct permission.
                    elif existing_permission.permission < permission_model.permission:
                        getattr(existing_permission, attribute_name).remove(entity)
                getattr(new_permission_model, attribute_name).add(entity)


def fetch_user(query: str) -> User:
    """Get user by ``pk``, ``username`` or ``email``.

    :raises ParseError: if user can not be determined.
    """
    user_filter = models.Q(username=query) | models.Q(email=query)
    if query.isdigit():
        user_filter |= models.Q(pk=query)

    user_model = get_user_model()
    try:
        return user_model.objects.get(user_filter)
    except user_model.DoesNotExist:
        raise exceptions.ParseError("Unknown user: {}".format(query))
    except user_model.MultipleObjectsReturned:
        raise exceptions.ParseError("Cannot uniquely determine user: {}".format(query))


def fetch_group(query: str) -> Group:
    """Get group by ``pk`` or ``name``. Raise error if it doesn't exist."""
    group_filter = {"pk": query} if query.isdigit() else {"name": query}

    try:
        return Group.objects.get(**group_filter)
    except Group.DoesNotExist:
        raise exceptions.ParseError("Unknown group: {}".format(query))


def check_owner_permission(payload: dict, allow_user_owner: bool):
    """Raise ``PermissionDenied``if ``owner`` found in ``data``."""
    for entity_type in ["users", "groups"]:
        for permission in payload.get(entity_type, {}).values():
            if permission == "owner":
                if entity_type == "users" and allow_user_owner:
                    continue

                if entity_type == "groups":
                    raise exceptions.ParseError(
                        "Owner permission cannot be assigned to a group"
                    )

                raise exceptions.PermissionDenied(
                    "Only owners can grant/revoke owner permission"
                )


def check_public_permissions(payload: dict):
    """Raise ``PermissionDenied`` if public permissions are too open."""
    allowed_public_permissions = ["view"]
    if "public" in payload:
        if payload["public"] not in allowed_public_permissions:
            raise exceptions.PermissionDenied(
                "Permissions for public users are too open"
            )


def check_user_permissions(payload: dict, user_pk: int):
    """Raise ``PermissionDenied`` if ``payload`` includes ``user_pk``."""
    user_pks = payload.get("users", {}).keys()
    if user_pk in user_pks:
        raise exceptions.PermissionDenied("You cannot change your own permissions")


def update_permission(obj: models.Model, data):
    """Update object permissions."""
    full_permissions = get_all_perms(obj)

    def apply_perm(permission: str, entity: Union[User, Group]):
        """Set permission on the object obj to the given entity.

        If given permission does not exist, ``exceptions.ParseError`` is
        raised.

        If special keyword "ALL" passed as ``permission`` top-level permission
        is set on the given object for the given entity.

        If special keywore "NONE" is passed as ``permission`` all permissions
        are revoked.

        :param entity: user or group to set permissions to.
        """
        if permission == "NONE":
            obj.set_permission(None, entity)
        elif permission == "ALL":
            for current_perm in reversed(
                Permission.highest_permission().all_permission_objects()
            ):
                if str(permission) in full_permissions:
                    obj.set_permission(current_perm.value, entity)
                    break
            raise RuntimeError("Object must have at least one permission.")
        else:
            if permission not in full_permissions:
                raise exceptions.ParseError("Unknown permission: {}".format(permission))
            obj.set_permission(permission, entity)

    def set_permissions(entity_type):
        """Set object permissions."""
        fetch_entity = fetch_user if entity_type == "users" else fetch_group

        for entity_id in data.get(entity_type, {}):
            apply_perm(data[entity_type][entity_id], fetch_entity(entity_id))

    def set_public_permissions():
        """Set public permissions."""
        if "public" in data:
            apply_perm(data["public"], get_anonymous_user())

    with transaction.atomic():
        set_permissions("users")
        set_permissions("groups")
        set_public_permissions()


def assign_contributor_permissions(obj, contributor=None):
    """Assign all permissions to object's contributor."""
    permission = str(Permission.highest_permission())
    set_permission(permission, contributor or obj.contributor, obj)



"""Custom permissions for Flow API."""
from typing import List, Optional

from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.http import Http404, HttpRequest

from rest_framework import permissions

from resolwe.permissions.models import Permission
from resolwe.permissions.utils import get_anonymous_user, user_model_check


def _permission_check(permission_name: str, obj: models.Model) -> str:
    """Check the app labels on permission and object.

    :raises RuntimeError: when the permission app name and model app name
        differ.
    """
    # Check if app labels match.
    # TODO: I do not understand why the second part of check is necessary.
    if permission_name is not None and "." in permission_name:
        app_label, base_permission_name = permission_name.split(".", 1)
        if app_label != obj._meta.app_label:
            # Check the content_type app_label when permission
            # and obj app labels don't match.
            ctype = ContentType.objects.get_for_model(obj)
            if app_label != ctype.app_label:
                # TODO: custom error.
                raise RuntimeError(
                    "Passed perm has app label of '%s' while "
                    "given obj has app label '%s' and given obj"
                    "content_type has app label '%s'"
                    % (app_label, obj._meta.app_label, ctype.app_label)
                )
        return base_permission_name
    return permission_name


class ResolwePermissionBackend:
    """Custom resolwe permission backend."""

    supports_object_permissions = True
    supports_anonymous_user = True
    supports_inactive_user = True

    def authenticate(
        self,
        request: HttpRequest,
        username: Optional[User] = None,
        password: Optional[str] = None,
    ):
        """Authenticate user.

        Since our backend is only used for authorization return None. For
        details see customizing authentication in Django.

        https://docs.djangoproject.com/en/3.2/topics/auth/customizing/
        """
        return None

    def has_perm(
        self, user_obj: User, perm: str, obj: Optional[models.Model] = None
    ) -> bool:
        """
        Return ``True`` if given ``user_obj`` has ``perm`` for ``obj``.

        If no ``obj`` is given, ``False`` is returned.

        .. note::

           Remember, that if user is not *active*, all checks would return
           ``False``.

        Main difference between Django's ``ModelBackend`` is that we can pass
        ``obj`` instance here and ``perm`` doesn't have to contain
        ``app_label`` as it can be retrieved from given ``obj``.

        **Inactive user support**

        If user is authenticated but inactive at the same time, all checks
        always returns ``False``.
        """
        should_check, user_obj, obj = user_model_check(user_obj, obj)
        if not should_check:
            return False

        perm = _permission_check(perm, obj)
        required_permission = Permission.from_name(perm).value

        assert hasattr(
            obj, "permission_group"
        ), f"Object of type {obj._meta.label} has no permission group field."
        userperm_queryset = obj._meta.model.objects.filter(
            pk=obj.pk,
            permission_group__permissions__users=user_obj,
            permission_group__permissions__permission__gte=required_permission,
        )
        groupperm_queryset = obj._meta.model.objects.filter(
            pk=obj.pk,
            permission_group__permissions__groups__in=user_obj.groups.all(),
            permission_group__permissions__permission__gte=required_permission,
        )
        return userperm_queryset.exists() or groupperm_queryset.exists()

    def get_all_permissions(
        self, user_obj: User, obj: Optional[models.Model] = None
    ) -> List[str]:
        """Return a set of permissions for ``user_obj`` on ``obj``."""
        should_check, user_obj, obj = user_model_check(user_obj, obj)
        if not should_check:
            return []
        permission_value = obj.permission_group.permissions.filter(
            models.Q(users=user_obj) | models.Q(groups__in=user_obj.groups.all())
        ).aggregate(permission=models.Max("permission"))["permission"]
        if permission_value:
            return Permission(permission_value).all_permissions()
        else:
            return []


class ResolwePermissions(permissions.DjangoObjectPermissions):
    """Resolwe Permissions."""

    perms_map = {
        "GET": ["%(app_label)s.view"],
        "OPTIONS": ["%(app_label)s.view"],
        "HEAD": ["%(app_label)s.view"],
        "POST": ["%(app_label)s.edit"],
        "PUT": ["%(app_label)s.edit"],
        "PATCH": ["%(app_label)s.edit"],
        "DELETE": ["%(app_label)s.edit"],
    }

    def has_permission(self, request, view):
        """Return `True` as we don't use model level perms."""
        return True

    def has_object_permission(self, request, view, obj):
        """Check object permissions."""
        # admins can do anything
        if request.user.is_superuser:
            return True

        # `share` permission is required for editing permissions
        if "permissions" in view.action:
            self.perms_map["POST"] = ["%(app_label)s.share"]

        if hasattr(view, "get_queryset"):
            queryset = view.get_queryset()
        else:
            queryset = getattr(view, "queryset", None)

        assert queryset is not None, (
            "Cannot apply DjangoObjectPermissions on a view that "
            "does not set `.queryset` or have a `.get_queryset()` method."
        )

        model_cls = queryset.model
        user = request.user
        # Return permissions based on the method and perms_map above. Since our
        # permissions are linearly ordered and lower are includad into higher
        # this must always only be highest possible permission.
        # Searching for more than one permission will slow things down.
        perms = self.get_required_object_permissions(request.method, model_cls)

        if not user.has_perms(perms, obj) and not get_anonymous_user().has_perms(
            perms, obj
        ):
            if request.method not in permissions.SAFE_METHODS:
                # If the user does not have permissions we need to determine if
                # they have read permissions to see 403, or not, and simply see
                # a 404 response.
                read_perms = self.get_required_object_permissions("GET", model_cls)
                if user.has_perms(read_perms, obj):
                    return False

            raise Http404

        return True



# pylint: disable=missing-docstring

from resolwe.flow.models import Collection, DescriptorSchema, Process
from resolwe.permissions.utils import copy_permissions, get_perms, set_permission
from resolwe.test import TestCase


class UtilsTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.src_process = Process.objects.create(
            name="Source process", contributor=self.contributor
        )
        self.dst_process = Process.objects.create(
            name="Destination process", contributor=self.contributor
        )

        self.collection = Collection.objects.create(
            name="Test collection", contributor=self.contributor
        )

        self.descriptor_schema = DescriptorSchema.objects.create(
            contributor=self.contributor
        )

    def test_copy_permissions(self):
        self.src_process.set_permission("view", self.contributor)
        self.src_process.set_permission("view", self.group)

        copy_permissions(self.src_process, self.dst_process)
        self.assertTrue(self.contributor.has_perm("flow.view", self.dst_process))
        # User inherites permission from group
        self.assertTrue(self.user.has_perm("flow.view", self.dst_process))

    def test_copy_different_ctype(self):
        # TODO: this could be annoying. Even assigning 'edit' to process might work.

        set_permission("edit", self.contributor, self.collection)

        copy_permissions(self.collection, self.dst_process)

        self.assertEqual(get_perms(self.contributor, self.dst_process), ["view"])
